//**************************************************************************/
//*File Integrate2.c
//*
//*Made the Aug 16 14:23 1995 by Nicolas Piller
//*
//*
//*Desctiption: Defininition of a classes to solve numericaly integrals.
//             
//*
//*Change: Date                       by
//*        ...                        ...
//**************************************************************************/
#include <iostream.h>
#include "LibNic.h"
#include "Integrate.h"
#include "Integrate.f.h"

//**************************************************************************/

typedef void(*DINTGK)(double(*)(double*,double*,void*),int*,double*,int*,
                      double*,double*,int*,int*,double*,double*,int*,int*,void*);

DINTGK GetDINTGK(int Level);
DINTGK GetDINTGK(int Level)

{ DINTGK P=0;
  switch (Level)
    {
    case 0: {P=&FortranCall(dintgk1a); break;}
    case 1: {P=&FortranCall(dintgk1b); break;}
    case 2: {P=&FortranCall(dintgk1c); break;}
    case 3: {P=&FortranCall(dintgk1d); break;}
    case 4: {P=&FortranCall(dintgk1e); break;}
    default: {FatalError("Integration: Unsuported Level");}
    }
  return P;
}

//**************************************************************************/

IntegrationDatas::IntegrationDatas()

{ //AccRel=0.000001;
  AccRel=1E-6;
  Pt=3;
  //Fcm=10000;
  Fcm=100000;
  Output=1;
} 

IntegrationDatas::IntegrationDatas(double accrel, int pt, int fcm)

{ AccRel=accrel;
  Pt=pt;
  Fcm=fcm;
  Output=1;
} 

IntegrationDatas::IntegrationDatas(double accrel, int pt, int fcm, int outputt)

{ AccRel=accrel;
  Pt=pt;
  Fcm=fcm;
  Output=outputt;
} 

IntegrationDatas::IntegrationDatas(const IntegrationDatas& id)

{ AccRel=id.AccRel;
  Pt=id.Pt;
  Fcm=id.Fcm;
  Output=id.Output;
} 

IntegrationDatas::PrintError(int OK)

{  if(Output==0) return 0;
   if((Output==1)&&(OK!=0)) cerr<<"IE("<<OK<<") ";
   if((Output==2)&&(OK!=0))
     {
     switch(OK)
       {
       case(1): {cerr<<"IE()";    break;}
       case(2): {cerr<<"IE()";    break;}
       case(3): {cerr<<"IE()";    break;}
       case(4): {cerr<<"IE()";    break;}
       case(5): {cerr<<"IE()";    break;}
       }
    }
  if((Output!=0)&&(Output!=1)&&(Output!=2)) return -1;
  return 0;
}

//**************************************************************************/

Integrate1d::Integrate1d(int Nf)

{ level=0;
  NFun=Nf;
  if (Nf!=0)TmpRes= new double [NFun+1];
}

Integrate1d::Integrate1d(int Nf,IntegrationDatas id)

{ level=0;
  NFun=Nf;
  TmpRes= new double [NFun+1];
  Datas=id; 
}

void Integrate1d::SetLevel(int lev)

{ if((lev<0)||(lev>4)) FatalError("Integration: Unsuported Level");
  level=lev;
}

Integrate1d::~Integrate1d()

{ delete TmpRes;
}


int Integrate1d::Integrate(double *Result,
                           Object *O,void(Object:: *FObj)(double*,double),
                              double a,double b)

{ OF=O;
  MF=FObj;
  F=FUN;
  int IVar=1;
  double var[1];
  DINTGK fintegr=GetDINTGK(level);

  (*fintegr)((double(*)(double*, double*, void*))F,
          &NFun,var,&IVar,&a,&b,&Datas.Pt,&Datas.Fcm,&Datas.AccRel,TmpRes,
          &M,&IOK,(void*)this);

  for(int i=0;i<NFun;i++) Result[i]= TmpRes[i+1];
  Datas.PrintError(IOK);
  return IOK;  
}

int Integrate1d::IntegratePInf(double *Result,
                               Object *O,void(Object:: *FObj)(double*,double),
                               double a,double RenPos)

{ if (a>RenPos)
    { 
    cerr<<"(Integrate1dPInf) Error: b<RenPos\n";
    RenPos=2*a-RenPos;
    }
  OF=O;
  MF=FObj;
  F=FUNPInf;
  int IVar=1;
  double var[1];
  LimN=a;
  RenP=RenPos;
  double A=0;
  double B=1.67;
  DINTGK fintegr=GetDINTGK(level);

  (*fintegr)((double(*)(double*, double*, void*))F,
           &NFun,var,&IVar,&A,&B,&Datas.Pt,&Datas.Fcm,&Datas.AccRel,TmpRes,
           &M,&IOK,(void*)this);
  for(int i=0;i<NFun;i++) Result[i]= TmpRes[i+1];
  Datas.PrintError(IOK);
  return IOK;  
}

int Integrate1d::IntegrateNInf(double *Result,
                               Object *O,void(Object:: *FObj)(double*,double),
                               double b,double RenNeg)

{ if (b<RenNeg)
    { 
    cerr<<"(Integrate1dNInf) Error: b<RenNeg\n";
    RenNeg=2*b-RenNeg;
    }
  OF=O;
  MF=FObj;
  F=FUNNInf;
  int IVar=1;
  double var[1];
  LimP=b;
  RenN=RenNeg;
  double A=-1.67;
  double B=0;
  DINTGK fintegr=GetDINTGK(level);
  
  (*fintegr)((double(*)(double*, double*, void*))F,
           &NFun,var,&IVar,&A,&B,&Datas.Pt,&Datas.Fcm,&Datas.AccRel,TmpRes,
           &M,&IOK,(void*)this);
  for(int i=0;i<NFun;i++) Result[i]= TmpRes[i+1];
  Datas.PrintError(IOK);
  return IOK;  
}

int Integrate1d::IntegrateNPInf(double *Result,
                                Object *O,void(Object:: *FObj)(double*,double),
                                double RenNeg,double RenPos)

{ if (RenNeg>RenPos)
    { 
    cerr<<"(Integrate1dNPInf) Error: RenNeg>RenPos\n";
    double tmp=RenNeg; RenNeg=RenPos; RenPos=tmp;
    }

  OF=O;
  MF=FObj;
  F=FUNNPInf;
  int IVar=1;
  double var[1];
  RenP=RenPos;
  RenN=RenNeg;
  double A=-1.67;
  double B=+1.67;
  DINTGK fintegr=GetDINTGK(level);

 (*fintegr)((double(*)(double*, double*, void*))F,
           &NFun,var,&IVar,&A,&B,&Datas.Pt,&Datas.Fcm,&Datas.AccRel,TmpRes,
           &M,&IOK,(void*)this);
  for(int i=0;i<NFun;i++) Result[i]= TmpRes[i+1];
  Datas.PrintError(IOK);
  return IOK;  
}

void Integrate1d::FctValue(double* Res,double var)

{ (OF->*MF)(Res,var);
}

double FUN(double* var, double* Fc,Integrate1d* This)

{ double Tot=0;
  int i;

  This->FctValue(Fc,var[0]);  
  for(i=0;i<This->NFun;i++)
    {
    Tot=Tot+Fc[i];
    }
  return Tot;
}

double FUNPInf(double* var, double* Fc,Integrate1d* This)

{ double Tot=0;
  double upref = (This->RenP-This->LimN);
  double up    = upref*tan(var[0]);
  double dup   = upref/(cos(var[0])*cos(var[0]));
  int i;

  This->FctValue(Fc,up+This->LimN);  
  for(i=0;i<This->NFun;i++)
    {    
    Fc[i]*=dup;
    Tot=Tot+Fc[i];
    }
  return Tot;
}

double FUNNInf(double* var, double* Fc,Integrate1d* This)

{ double Tot=0;
  double upref = (This->RenN-This->LimP);
  double up    = -upref*tan(var[0]);
  double dup   = -upref/(cos(var[0])*cos(var[0]));

  This->FctValue(Fc,up+This->LimN);  
  for(int i=0;i<This->NFun;i++)
    {    
    Fc[i]*=dup;
    Tot=Tot+Fc[i];
    }
  return Tot;
}

double FUNNPInf(double* var, double* Fc,Integrate1d* This)

{ double Tot=0;
  double upref = (This->RenP-This->RenN)/2.0;
  double up    = upref*tan(var[0]);
  double dup   = upref/(cos(var[0])*cos(var[0]));

  This->FctValue(Fc,up+(This->RenP+This->RenN)/2.0);  
  for(int i=0;i<This->NFun;i++)
    {    
    Fc[i]*=dup;
    Tot=Tot+Fc[i];
    }
  return Tot;
}

Integrate1dnd::Integrate1dnd()
              : Integrate1d(0)

{ F=FUN;
}

Integrate1dnd::Integrate1dnd(int NFc)
              : Integrate1d(NFc)
{ F=FUN;
}


void Integrate1dnd::SetConst(int Level_,IntegrationDatas id, double *Var_)
                
{ Level=Level_;
  Var=Var_;
  Datas=id;
}

void Integrate1dnd::SetFunction(Object *O,void(Object:: *FObj)(double*,double))

{ OF=O;
  MF=FObj;
}
				    

void Integrate1dnd::IntegrateValue(double* Result,double X)

{ int IVar=1;
  double var[1];
  if (Level>0) Var[Level-1]=X;
  DINTGK fintegr=GetDINTGK(Level);

  (*fintegr)((double(*)(double*, double*, void*))F,
           &NFun,var,&IVar,&a,&b,&Datas.Pt,&Datas.Fcm,&Datas.AccRel,TmpRes,
           &M,&IOK,(void*)this);

  for (int i=0;i<NFun;i++)Result[i]=TmpRes[i+1];
  Datas.PrintError(IOK);
}


IntegrateNd::IntegrateNd(int N_,int NF)

{ int i;
  N=N_;
  IntegrL = new Integrate1dnd [N]();
  
  VarNd = new double[N];

  IntegrationDatas id=IntegrationDatas();

  for (i=0;i<N;i++)
    {
    IntegrL[i]=Integrate1dnd(NF);
    IntegrL[i].SetConst(i, id, VarNd);
    }
}

IntegrateNd::~IntegrateNd()

{ delete IntegrL;
  delete VarNd;
}

IntegrateNd::IntegrateNd(int N_,int NF,IntegrationDatas id[])

{ int i;
  N=N_;
  IntegrL = new Integrate1dnd [N] ();
  VarNd = new double[N];

  for (i=0;i<N;i++)
    {    
    IntegrL[i]=Integrate1dnd(NF);
    IntegrL[i].SetConst(i, id[i],VarNd);
    }
}

void IntegrateNd::FctValue(double* Result,double X)

{  VarNd[N-1]=X;
   (OF->*MF)(Result,VarNd);
}


int IntegrateNd::Integrate(double* Result,Object *O,
                           void (Object:: *FObj)(double*, double *),
                              double A[],double B[])

{ int i;

  OF=O;
  MF=FObj;
  for(i=0;i<N-1;i++)
    { 
    IntegrL[i].SetLimits(A[i],B[i]);
    IntegrL[i].SetFunction((Object*)&IntegrL[i+1],
          (void(Object::*)(double*, double ))&Integrate1dnd::IntegrateValue);
    }
  IntegrL[N-1].SetLimits(A[N-1],B[N-1]);
  IntegrL[N-1].SetFunction(this,
            (void(Object::*)(double*, double ))&IntegrateNd::FctValue);

  IntegrL[0].IntegrateValue(Result,0);
  return OK();
}


int IntegrateNd::OK()

{ int ok=0, i;
  for (i=0;i<N;i++) ok=ok|IntegrL[i].OK();
  return ok;
}

int IntegrateNd::OK(int OkArray[])

{ int i, ok=0;
  for (i=0;i<N;i++) 
    {
    ok=ok|IntegrL[i].OK();
    OkArray[i]=IntegrL[i].OK();
    }
  return ok;
}
  



#define INTEGRATESRC(TN,TF,NFtc)                                          \
TF Integrate1d##TN::Integrate(Object *O,TF (Object:: *FObj)(double),      \
                                 double a,double b)                       \
                                                                          \
{ TF Result;                                                              \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::Integrate((double*)&Result,(Object *)this,                 \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
TF Integrate1d##TN::IntegratePInf(Object *O,TF (Object:: *FObj)(double),  \
                                 double a,double b)                       \
                                                                          \
{ TF Result;                                                              \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::IntegratePInf((double*)&Result,(Object *)this,             \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
TF Integrate1d##TN::IntegrateNInf(Object *O,TF (Object:: *FObj)(double),  \
                                 double a,double b)                       \
                                                                          \
{ TF Result;                                                              \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::IntegrateNInf((double*)&Result,(Object *)this,             \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
TF Integrate1d##TN::IntegrateNPInf(Object *O,TF (Object:: *FObj)(double)  \
                                 ,double a,double b)                      \
                                                                          \
{ TF Result;                                                              \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::IntegrateNPInf((double*)&Result,(Object *)this,            \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
                                                                          \
TF IntegrateNd##TN##::Integrate(Object *O,TF (Object:: *FObj)(double*),   \
                                 double* a,double* b)                     \
                                                                          \
{ TF Result;                                                              \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  IntegrateNd::Integrate((double*)&Result,(Object *)this,                 \
    (void (Object:: *)(double*,double*))IntegrateNd##TN::AFctValue,a,b);  \
  return Result;                                                          \
}                                                                         \
                                                                          \
void Integrate1d##TN::AFctValue(double *Res,double Var)                   \
{ TF C=(COF->*CMF)(Var);                                                  \
  (*(TF*)Res)=C;                                                          \
};                                                                        \
                                                                          \
void IntegrateNd##TN::AFctValue(double *Res,double* Var)                  \
                                                                          \
{ TF C=(COF->*CMF)(Var);                                                  \
  (*(TF*)Res)=C;                                                          \
};                                                                        \



#define INTEGRATESRCMATRIX(TN,TF,NF_)     TF                              \
Integrate1d##TN::Integrate(Object *O,TF (Object:: *FObj)(double),         \
                                 double a,double b)                       \
{ TF Result(I,J);                                                         \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::Integrate((double*)Result.array(),(Object *)this,          \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
TF Integrate1d##TN::IntegratePInf(Object *O,TF (Object:: *FObj)(double),  \
                                 double a,double b)                       \
{ TF Result(I,J);                                                         \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::IntegratePInf((double*)Result.array(),(Object *)this,      \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
TF Integrate1d##TN::IntegrateNInf(Object *O,TF (Object:: *FObj)(double),  \
                                 double a,double b)                       \
{ TF Result(I,J);                                                         \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::IntegrateNInf((double*)Result.array(),(Object *)this,      \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
TF Integrate1d##TN::IntegrateNPInf(Object *O,TF (Object:: *FObj)(double), \
                                 double a,double b)                       \
{ TF Result(I,J);                                                         \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  Integrate1d::IntegrateNPInf((double*)Result.array(),(Object *)this,     \
    (void (Object:: *)(double*,double))Integrate1d##TN::AFctValue,a,b);   \
  return Result;                                                          \
}                                                                         \
                                                                          \
void Integrate1d##TN::AFctValue(double *Res,double Var)                   \
                                                                          \
{ TF C=(COF->*CMF)(Var);                                                  \
  if((C.size().I!=I)||(C.size().J!=J))                                    \
    FatalError("Matrix size uncompatible for Integration");               \
  for(int i=0;i<NFun;i++)                                                 \
  Res[i]=((double*)C.array())[i];                                         \
};                                                                        \
                                                                          \
TF IntegrateNd##TN::Integrate(Object *O,TF (Object:: *FObj)(double*),     \
                                 double* a,double* b)                     \
                                                                          \
{ TF Result;                                                              \
  COF=O;                                                                  \
  CMF=FObj;                                                               \
  IntegrateNd::Integrate((double*)Result.array(),(Object *)this,          \
    (void (Object:: *)(double*,double*))IntegrateNd##TN::AFctValue,a,b);  \
  return Result;                                                          \
}                                                                         \
                                                                          \
void IntegrateNd##TN::AFctValue(double *Res,double* Var)                  \
                                                                          \
{ TF C=(COF->*CMF)(Var);                                                  \
  if((C.size().I!=I)||(C.size().J!=J))                                    \
    FatalError("Matrix size uncompatible for Integration");               \
  for(int i=0;i<I*J*NF_;i++)                                              \
  Res[i]=((double*)C.array())[i];                                         \
};                                                                        \


INTEGRATESRC (Complex,Complex,2)
INTEGRATESRC (Double,double,1)
INTEGRATESRC (Vect2d,vect2d,2)
INTEGRATESRC (Vect3d,vect3d,3)
INTEGRATESRC (Vect2dC,vect2dc,4)
INTEGRATESRC (Vect3dC,vect3dc,6)

//INTEGRATESRCMATRIX(DMatrix,dmatrix,1)
//INTEGRATESRCMATRIX(CMatrix,cmatrix,2)

